#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-12
# @Filename: models.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from .base import Extension, FITSModel, HeaderModel


#: A basic header model with camera and exposure information.
basic_header_model = HeaderModel(
    [
        'VCAM',
        'CAMNAME',
        'CAMUID',
        'IMAGETYP',
        'EXPTIME',
        'EXPTIMEN',
        'STACK',
        'STACKFUN',
        ('TIMESYS', 'TAI', 'The time scale system'),
        ('DATE-OBS', '{__exposure__.obstime.tai.isot}', 'Date (in TIMESYS) the exposure started')
    ]
)


#: A basic FITS model for uncompressed images. Includes a single extension
#: with the raw data and a `.basic_header_model`.
basic_fits_model = FITSModel(
    [Extension(data='raw',
               header_model=basic_header_model,
               name='PRIMARY')]
)


#: A compressed, basic FITS model. Similar to `.basic_fits_model` but uses
#: ``RICE_1`` compression.
basic_fz_fits_model = FITSModel(
    [Extension(data='raw',
               header_model=basic_header_model,
               compressed='RICE_1',
               name='PRIMARY')]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-12
# @Filename: magic.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


_MAGIC_CARDS = {

    'EXPTIME': ('{__exposure__.exptime}', 'Exposure time of single integration [s]'),
    'EXPTIMEN': ('{__exposure__.exptime_n}', 'Total exposure time [s]'),
    'STACK': ('{__exposure__.stack}', 'Number of stacked frames'),
    'STACKFUN': ('{__exposure__.stack_function}', 'Function used for stacking'),
    'OBSTIME': ('{__exposure__.obstime.tai}', 'Time of the start of the exposure [TAI]'),
    'IMAGETYP': ('{__exposure__.image_type}', 'The image type of the file'),
    'CAMNAME': ('{__camera__.name}', 'Camera name'),
    'CAMUID': ('{__camera__.uid}', 'Camera UID'),
    'VCAM': ('{__camera__.__version__}', 'The version of camera library')

}



# encoding: utf-8
# flake8: noqa

from sdsstools import get_package_version

from .camera import *
from .events import *
from .exceptions import *
from .exposure import *
from .notifier import *


NAME = 'sdss-basecam'

__version__ = get_package_version(__file__, 'sdss-basecam') or 'dev'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-11
# @Filename: camera.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import os

import click

from clu.parser import CluCommand, CluGroup, help_, ping

from ..events import CameraEvent
from ..exceptions import CameraConnectionError, CameraError, ExposureError
from .tools import get_cameras


@click.group(cls=CluGroup)
def camera_parser():
    pass


camera_parser.add_command(help_)
camera_parser.add_command(ping)


@camera_parser.command(name='list')
async def list_(command):
    """Lists cameras connected to the camera system."""

    cameras = [camera.name for camera in command.actor.camera_system.cameras]
    command.finish(cameras=cameras, default_cameras=command.actor.default_cameras)


@camera_parser.command()
@click.argument('CAMERAS', nargs=-1, type=str)
@click.option('-f', '--force', is_flag=True, default=False,
              help='Forces a camera to be set as default '
                   'even if it is not connected.')
async def set_default(command, cameras, force):
    """Set default cameras."""

    camera_system = command.actor.camera_system

    camera_names = []
    for camera in cameras:
        for camera_split in camera.split(','):
            if camera_split not in camera_names:
                camera_names.append(camera_split)

    for camera_name in camera_names:
        camera = camera_system.get_camera(name=camera_name)

        if not camera:
            msg = f'camera {camera_name} is not connected.'
            if force:
                command.write('w', text=msg)
            else:
                return command.fail(camera=camera.name,
                                    text=f'camera {camera_name} is not connected.')

    command.actor.set_default_cameras(camera_names)

    command.finish(default_cameras=command.actor.default_cameras)


@camera_parser.command()
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
async def status(command, cameras):
    """Returns the status of a camera."""

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    for camera in cameras:
        status = {'camera': {'name': camera.name,
                             'uid': camera.uid},
                  'status': camera.get_status(update=True)}
        command.info(status)

    command.finish()


@camera_parser.command()
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.option('--timeout', '-t', type=float, default=5., show_default=True,
              help='Seconds to wait until disconnect or reconnect command times out.')
async def reconnect(command, cameras, timeout):
    """Reconnects a camera."""

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        command.warning(text=f'reconnecting camera {camera.name!r}')

        try:
            await asyncio.wait_for(camera.disconnect(), timeout=timeout)
            command.info(camera=camera.name,
                         text=f'camera {camera.name!r} was disconnected.')
        except CameraConnectionError as ee:
            command.warning(camera=camera.name,
                            text=f'camera {camera.name!r} fail to disconnect: '
                                 f'{ee}. Will try to reconnect.')
        except asyncio.TimeoutError:
            command.warning(camera=camera.name,
                            text=f'camera {camera.name!r} timed out '
                                 'disconnecting. Will try to reconnect.')

        try:
            await asyncio.wait_for(camera.connect(force=True), timeout=timeout)
            command.info(camera=camera.name,
                         error=f'camera {camera.name!r} was reconnected.')
        except CameraConnectionError as ee:
            command.error(camera=camera.name,
                          error=f'camera {camera.name!r} fail to reconnect: {ee}')
            failed = True
        except asyncio.TimeoutError:
            command.error(camera=camera.name,
                          error=f'camera {camera.name!r} timed out reconnecting.')
            failed = True

    if failed:
        return command.fail('some cameras failed to reconnect.')
    else:
        return command.finish()


@camera_parser.command()
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.argument('EXPTIME', type=float, required=False)
@click.option('--object', 'image_type', flag_value='object', default=True,
              show_default=True, help='Takes an object exposure.')
@click.option('--flat', 'image_type', flag_value='flat',
              show_default=True, help='Takes a flat exposure.')
@click.option('--dark', 'image_type', flag_value='dark',
              show_default=True, help='Takes a dark exposure.')
@click.option('--bias', 'image_type', flag_value='bias',
              show_default=True, help='Takes a bias exposure.')
@click.option('--filename', '-f', type=str, default=None,
              show_default=True, help='Filename of the imaga to save.')
@click.option('--stack', '-s', type=int, default=1,
              show_default=True, help='Number of images to stack.')
async def expose(command, cameras, exptime, image_type, filename, stack):
    """Exposes and writes an image to disk."""

    def stage(event, payload):
        name = payload.get('name', None)
        if event in CameraEvent:
            stage = event.value
            command.info(name=name, stage=stage)

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    if image_type == 'bias':
        if exptime and exptime > 0:
            command.warning('seeting exposure time for bias to 0 seconds.')
        exptime = 0.0

    if filename and len(cameras) > 1:
        return command.fail('--filename can only be used when exposing '
                            'a single camera.')

    failed = False
    for camera in cameras:

        command.actor.listener.register_callback(stage)
        command.info(text=f'exposing camera {camera.name!r}')

        try:
            # Schedule camera.expose as a task to allow the events to be
            # notified concurrently.
            exposure = await command.actor.loop.create_task(
                camera.expose(exptime, image_type=image_type, stack=stack,
                              filename=filename, write=True))
            file_path = os.path.realpath(str(exposure.filename))
            command.info(camera=camera.name, filename=file_path)
        except ExposureError as ee:
            command.error(camera=camera.name,
                          error='error found while exposing camera '
                                f'{camera.name!r}: {ee!s}')
            failed = True
        finally:
            command.actor.listener.remove_callback(stage)

    if failed:
        return command.failed('one or more cameras failed to expose.')
    else:
        return command.finish()


@click.command(cls=CluCommand)
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.option('--open', 'shutter_position', flag_value='open',
              help='Open the shutter.')
@click.option('--close', 'shutter_position', flag_value='close',
              help='Close the shutter.')
async def shutter(command, cameras, shutter_position):
    """Controls the camera shutter.

    If called without a shutter position flag, returns the current position
    of the shutter.

    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:
        if shutter_position is None:
            shutter_now = await camera.get_shutter()
            command.info(camera=camera.name, shutter=shutter_now)
        else:
            try:
                await camera.set_shutter(shutter_position == 'open')
                command.info(camera=camera.name, shutter=shutter_position)
            except CameraError as ee:
                command.error(camera=camera.name,
                              error=f'failed commanding shutter: {ee!s}')
                failed = True

    if failed:
        return command.fail('failed to command the shutter of one or more cameras.')
    else:
        return command.finish()


@click.command(cls=CluCommand)
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.argument('TEMPERATURE', type=float, required=False)
async def temperature(command, cameras, temperature):
    """Controls the camera temperature.

    If called without a temperature value, returns the current temperature.

    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    temperature_tasks = []

    for camera in cameras:

        if not temperature:
            command.info(camera=camera.name,
                         temperature=await camera.get_temperature())
        else:
            temperature_tasks.append(camera.set_temperature(temperature))

    if not temperature:
        return command.finish()

    results = await asyncio.gather(*temperature_tasks, return_exceptions=True)
    failed = False
    for ii, result in enumerate(results):
        if isinstance(result, CameraError):
            command.error(camera=cameras[ii].name,
                          error=str(result))
            failed = True

    if failed:
        return command.fail('one or more cameras failed to set temperature.')
    else:
        return command.finish('all cameras have reached their set points.')


@click.command(cls=CluCommand)
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.argument('BINNING', nargs=2, type=int, required=False)
async def binning(command, cameras, binning):
    """Controls the camera binning.

    If called without a binning value, returns the current value.

    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        if not binning:
            command.info(camera=camera.name,
                         binning=tuple(await camera.get_binning()))
        else:
            try:
                await camera.set_binning(*binning)
                command.info(camera=camera.name,
                             binning=tuple(binning))
            except (CameraError, AssertionError) as ee:
                command.error(camera=camera.name,
                              error=str(ee))
                failed = True

    if failed:
        return command.fail('failed to set binning for one or more cameras.')
    else:
        return command.finish()


@click.command(cls=CluCommand)
@click.argument('CAMERAS', nargs=-1, type=str, required=False)
@click.argument('AREA', nargs=4, type=int, required=False)
@click.option('--reset', '-r', is_flag=True, default=False,
              show_default=True, help='Restores the original image area.')
async def area(command, cameras, area, reset):
    """Controls the camera image area.

    If called without an image area value, returns the current value.
    The image area must have the format (x0, x1, y0, y1) and be 1-indexed.

    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        if not area and reset is False:
            command.info(camera=camera.name,
                         area=tuple(await camera.get_image_area()))
        else:
            if reset:
                area = None
            try:
                await camera.set_image_area(area)
                command.info(camera=camera.name,
                             area=tuple(await camera.get_image_area()))
            except CameraError as ee:
                command.error(camera=camera.name,
                              error=str(ee))
                failed = True

    if failed:
        return command.fail('failed to set binning for one or more cameras.')
    else:
        return command.finish()


_MIXIN_TO_COMMANDS = {
    'ShutterMixIn': [shutter],
    'CoolerMixIn': [temperature],
    'ImageAreaMixIn': [binning, area]
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-03
# @Filename: events.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import enum


__all__ = ['CameraSystemEvent', 'CameraEvent']


class CameraSystemEvent(enum.Enum):
    """Enumeration of camera system events."""

    CAMERA_ADDED = enum.auto()
    CAMERA_REMOVED = enum.auto()


class CameraEvent(enum.Enum):
    """Enumeration of camera events."""

    CAMERA_CONNECTED = 'connected'
    CAMERA_CONNECT_FAILED = 'connect_failed'
    CAMERA_DISCONNECTED = 'disconnected'
    CAMERA_DISCONNECT_FAILED = 'disconnect_failed'

    EXPOSURE_FLUSHING = 'flushing'
    EXPOSURE_INTEGRATING = 'integrating'
    EXPOSURE_READING = 'reading'
    EXPOSURE_READ = 'read'
    EXPOSURE_FAILED = 'failed'
    EXPOSURE_WRITING = 'writing'
    EXPOSURE_WRITTEN = 'written'

    NEW_SET_POINT = 'new_set_point'
    SET_POINT_REACHED = 'set_point_reached'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-10
# @Filename: exposure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import functools
import os
import pathlib
import re

import astropy

from .exceptions import ExposureError
from .models import FITSModel


__all__ = ['Exposure', 'ImageNamer']


class Exposure(object):
    """Exposure class. Represents an exposure data and its metadata.

    An `.Exposure` is defined by the raw image taken by the camera, a series
    of attributes that define the exposure (exposure time, shutter, etc.) and
    a data model that is used to generate the FITS file for the exposure.

    Parameters
    ----------
    camera : .BaseCamera
        The instance of a subclass of `.BaseCamera` that took this exposure.
    filename : str
        The filename of the FITS image generated.
    data : numpy.array
        The exposure raw data, as a 2D array.
    fits_model : `.FITSModel`
        The `model <.FITSModel>` to create the FITS image. If `None`, a single
        extension with a basic header will be used.

    Attributes
    ----------
    image_type : str
        The type of image, one of ``bias``, ``flat``, ``dark``, ``object``.
    exptime : float
        The exposure time, in seconds, of a single integration.
    exptime_n : float
        The total exposure time, in seconds. If the image is stacked, this is
        the total time, i.e., the sum of the exposure time of each stacked
        image.
    stack : int
        Number of exposures stacked.
    stack_function : str
        Name of the function used for stacking.
    filename : str
        The path where to write the image.

    """

    def __init__(self, camera, filename=None, data=None, fits_model=None):

        self.camera = camera
        self.data = data
        self.fits_model = fits_model
        self.filename = filename

        self._obstime = None

        self.exptime = None
        self.exptime_n = None
        self.stack = 1
        self.stack_function = None

        self.image_type = None

        self.obstime = astropy.time.Time.now()

    @property
    def obstime(self):
        """The time at the beginning of the observation.

        It must be an `astropy.time.Time` object or a datetime in ISO
        format; in the latter case, UTC scale is assumed.

        """

        return self._obstime

    @obstime.setter
    def obstime(self, value):

        if isinstance(value, astropy.time.Time):
            self._obstime = value
        elif isinstance(value, str):
            self._obstime = astropy.time.Time(value, format='iso', scale='utc')
        else:
            raise ExposureError(f'invalid obstime {value}')

    def to_hdu(self, context={}):
        """Return an `~astropy.io.fits.HDUList` for the image.

        Parameters
        ----------
        context : dict
            A dictionary of arguments used to evaluate the FITS model.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            A list of HDUs in which the FITS data model has been evaluated
            for this exposure.

        """

        fits_model = self.fits_model

        if not fits_model:
            fits_model = FITSModel()

        return fits_model.to_hdu(self, context=context)

    async def write(self, filename=None, context={}, overwrite=False,
                    checksum=True):
        """Writes the image to disk.

        Parameters
        ----------
        filename : str
            The path where to write the file. If not provided, uses
            ``Exposure.filename``.
        context : dict
            A dictionary of arguments used to evaluate the FITS model.
        overwrite : bool
            Whether to overwrite the image if it exists.
        checksum : bool
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards to the
            headers of all HDUs written to the file.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            A list of HDUs in which the FITS data model has been evaluated
            for this exposure.

        """

        filename = filename or self.filename
        if not filename:
            raise ExposureError('filename not set.')

        hdulist = self.to_hdu(context=context)

        dirname = os.path.realpath(os.path.dirname(filename))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        writeto_partial = functools.partial(hdulist.writeto,
                                            filename,
                                            overwrite=overwrite,
                                            checksum=checksum)

        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, writeto_partial)

        return hdulist


class ImageNamer(object):
    """Creates a new sequential filename for an image.

    Parameters
    ----------
    basename : str
        The basename of the image filenames. Must contain a placeholder
        ``num`` in the place where to insert the sequence number. For example,
        ``'test-{num:04d}.fits'`` will produce image names ``test-0001.fits``,
        ``test-0002.fits``, etc. It's also possible to use placeholders for
        camera values, e.g. ``{camera.name}-{num}.fits``.
    dirname : str
        The directory for the images. Can include an expression based on the
        ``date`` substitution which is a `~astropy.time.Time.now` object. For
        example: ``dirname='/data/{camera.uid}/{int(date.mjd)}'``.
    overwrite : bool
        If `True`, the sequence will start at 1 regardless of the existing
        images. If `False`, the first element in the sequence will be selected
        to avoid colliding with any image already existing in the directory.
    camera : .BaseCamera
        A `.BaseCamera` instance. It can also be passed when calling the
        instance.

    Examples
    --------
    >>> namer = ImageNamer('{camera.name}-{num:04d}.fits', dirname='testdir')
    >>> namer(camera=camera)
    PosixPath('testdir/my_camera-0001.fits')
    >>> namer(camera=camera)
    PosixPath('testdir/my_camera-0002.fits')

    """

    def __init__(self, basename='{camera.name}-{num:04d}.fits',
                 dirname='.', overwrite=False, camera=None):

        assert re.match(r'.+(\{num.+\}).+', basename), 'invalid basename.'

        # We want to expand everything except the num first so we "double-escape" it.
        self.basename = re.sub(r'(\{num.+\})', r'{\1}', basename)
        self.dirname = pathlib.Path(dirname)
        self.overwrite = overwrite

        self._last_num = 0
        self.camera = camera

    def _eval_dirname(self):
        """Returns the evaluated dirname."""

        date = astropy.time.Time.now()

        return pathlib.Path(eval(f'f"{self.dirname}"', {}, {'date': date,
                                                            'camera': self.camera}))

    def _get_num(self, basename):
        """Returns the counter value."""

        if self.overwrite:
            return self._last_num + 1

        regex = re.compile(re.sub(r'\{num.+\}', '(?P<num>[0-9]*?)', basename))

        dirname = self._eval_dirname()
        all_files = list(map(str, dirname.glob('*')))

        match_files = list(filter(regex.search, all_files))

        if len(match_files) == 0:
            return self._last_num + 1

        values = [int(regex.search(file).group(1)) for file in match_files]
        return max(values) + 1

    def __call__(self, camera=None):

        camera = camera or self.camera

        if camera:
            expanded_basename = self.basename.format(camera=camera)
        else:
            expanded_basename = self.basename.format()

        num = self._get_num(expanded_basename)
        path = self._eval_dirname() / expanded_basename.format(num=num)
        self._last_num = num

        return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-10
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import abc
import functools
import warnings
from contextlib import contextmanager

import astropy.io.fits
import astropy.table
import numpy

from ..exceptions import CardError, CardWarning
from .magic import _MAGIC_CARDS


__all__ = ['FITSModel', 'Extension', 'HeaderModel', 'Card', 'CardGroup', 'MacroCard']


class FITSModel(list):
    """A model representing a FITS image.

    This model defines the data and header for each extension in a FITS
    image. The model can be evaluated for given `.Exposure`, returning a
    fully formed astropy `~astropy.io.fits.HDUList`.

    Parameters
    ----------
    extension : list of `.Extension`
        A list of HDU extensions defined as `.Extension` objects. If none is
        defined, a single, basic extension is added.

    """

    def __init__(self, extensions=None):

        extensions = extensions or []

        list.__init__(self, extensions)

        if len(self) == 0:
            self.append(Extension(data='raw', name='DATA'))

    def to_hdu(self, exposure, context={}):
        """Returns an `~astropy.io.fits.HDUList` from an exposure.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which the FITS model is evaluated.
        context : dict
            A dictionary of parameters used to fill the card replacement
            fields.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            A list of HDUs, each one define by its corresponding `.Extension`
            instance. The first extension is created as a
            `~astropy.io.fits.PrimaryHDU` unless it is compressed, in which
            case an empty primary HDU is prepended.

        """

        hdus = []

        for i, ext in enumerate(self):
            if i == 0:
                primary = True
                if ext.compressed:
                    # Compressed HDU cannot be primary.
                    hdus.append(astropy.io.fits.PrimaryHDU())
            else:
                primary = False
            hdus.append(ext.to_hdu(exposure, primary=primary, context=context))

        return astropy.io.fits.HDUList(hdus)


class Extension(object):
    """A model for a FITS extension.

    Parameters
    ----------
    data : str or numpy.ndarray
        The data for this FITS extension. Can be an array or a macro string
        indicating the type of data to store in the extension. Available macros
        are: ``'raw'`` for the raw image, or ``'none'`` for empty data. If
        `None`, the raw data will only be added to the primary HDU.
    header_model : .HeaderModel
        A `.HeaderModel` for this extension.
    name : str
        The name of the HDU.
    compressed : bool or str
        If `False`, the extension data will not be compressed. Otherwise, a
        string with one of the ``compression_type`` in
        `~astropy.io.fits.CompImageHDU`.

    """

    __VALID_DATA_VALUES = ['raw', 'none']

    def __init__(self, data=None, header_model=None, name=None, compressed=False):

        if isinstance(data, numpy.ndarray):
            self.data = data
        else:
            assert data is None or data in self.__VALID_DATA_VALUES, 'invalid data'
            self.data = data

        self.header_model = header_model

        self.name = name
        self.compressed = compressed

    def __repr__(self):

        return f'<Extension (name={self.name!r}, compressed={self.compressed})>'

    def to_hdu(self, exposure, primary=False, context={}):
        """Evaluates the extension as an HDU.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the extension.
        primary : bool
            Whether this is the primary extension.
        context : dict
            A dictionary of arguments used to evaluate the parameters in
            the extension.

        Returns
        -------
        hdu : `~astropy.io.fits.ImageHDU`
            An `~astropy.io.fits.ImageHDU` with the data and header evaluated
            for ``exposure``, or `~astropy.io.fits.CompImageHDU` if
            ``compressed=True``.

        """

        if self.compressed:
            HDUClass = astropy.io.fits.CompImageHDU
            if isinstance(self.compressed, str):
                HDUClass = functools.partial(HDUClass, compression_type=self.compressed)
        else:
            if primary:
                HDUClass = astropy.io.fits.PrimaryHDU
            else:
                HDUClass = astropy.io.fits.ImageHDU

        if not primary:
            HDUClass = functools.partial(HDUClass, name=self.name)

        data = self.get_data(exposure, primary=primary)

        # Create the HDU without a header first to allow astropy to create a
        # basic header (for example, if HDUClass is CompImageHDU this will add
        # the BITPIX keyword). Then append our header.
        hdu = HDUClass(data=data, header=None)

        if self.header_model:
            hdu.header.extend(self.header_model.to_header(exposure, context=context))

        return hdu

    def get_data(self, exposure, primary=False):
        """Returns the data as a numpy array."""

        if isinstance(self.data, str):
            if self.data == 'raw':
                data = exposure.data
            elif self.data == 'none':
                data = None
        elif self.data is None:
            data = exposure.data if primary else None
        else:
            data = self.data

        return data


class HeaderModel(list):
    """A model defining the header of an HDU.

    Parameters
    ----------
    cards : list
        A list of `.Card`, `.CardGroup`, or `MacroCard` instances. Elements
        can also be a tuple of two or three elements (for name, value, and
        optionally a comment).

    Examples
    --------
    >>> header_model = HeaderModel([('TELESCOP', 'APO-2.5', 'The telescope'),
                                    ('OBSERVATORY', 'APO'),
                                    Card('camname', '{(camera).name}')])


    """

    def __init__(self, cards=None):

        cards = [self._process_input(card) for card in cards or []]
        list.__init__(self, cards)

    def __repr__(self):

        return f'<HeaderModel {list.__repr__(self)!s}>'

    def _process_input(self, input_card):
        """Processes the input and converts it into a valid card."""

        if isinstance(input_card, (Card, CardGroup, MacroCard)):
            return input_card
        elif isinstance(input_card, (list, tuple)):
            if isinstance(input_card[0], (list, tuple)):
                card = CardGroup(input_card)
            else:
                card = Card(input_card)
            return card
        elif isinstance(input_card, str):
            return Card(input_card)
        else:
            raise CardError(f'invalid input {input_card!r}')

    def append(self, card):
        list.append(self, self._process_input(card))

    def insert(self, idx, card):
        list.insert(self, idx, self._process_input(card))

    def to_header(self, exposure, context={}):
        """Evaluates the header model for an exposure and returns a header.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the model.
        context : dict
            A dictionary of arguments used to evaluate the parameters in
            the model.

        Returns
        -------
        header : `~astropy.io.fits.Header`
            A `~astropy.io.fits.Header`, created by evaluating the model for
            the input exposure.

        """

        header = astropy.io.fits.Header()

        for card in self:

            if isinstance(card, Card):
                header.append(card.evaluate(exposure, context=context))
            elif isinstance(card, (CardGroup, MacroCard)):
                header += card.to_header(exposure, context=context)

        return header

    def describe(self):
        """Returns a table-like representation of the header model."""

        rows = []
        for card in self:
            if isinstance(card, Card):
                rows.append((card.name.upper(), card.value, card.comment))
            elif isinstance(card, CardGroup):
                for card_ in card:
                    rows.append((card_.name.upper(), card_.value, card_.comment))
            elif isinstance(card, MacroCard):
                rows.append(('### MACRO', card.__class__.__name__, ''))
            else:
                raise CardError(f'invalid card {card}. '
                                'This should not have happened.')

        return astropy.table.Table(rows=rows, names=['name', 'value', 'comment'])


class Card(object):
    """A card representing a single entry in the header.

    This class is similar to astropy's `~astropy.io.fits.Card` with the
    difference that the value of the card does not need to be a literal and is
    evaluated when `.evaluate` is called.

    The value of the card can be:

    - A Python literal.

    - A callable, with a list of arguments, that is called when the card
      is evaluated.

    - A string using Python's :ref:`python:formatstrings`. As with normal
      strings, the replacement fields are surrounded by ``{`` and ``}`` and
      filled when `.evaluate` is called. The replacement fields can point to an
      attribute, function, method, or property. In addition, it is possible to
      use the string ``__exposure__`` and ``__camera__`` to refer to the
      `.Exposure` instance for which this card will be evaluated, and the
      instance of `.BaseCamera` that took the image, respectively. For example,
      ``value='{__camera__.name}'`` will be replaced with the name of the
      camera, and ``value='{__exposure__.obstime}'`` will retrieve the time of
      the start of the exposure.

    - A string to be evaluated using Python's `eval` function. For example,
      ``value='__camera__.get_temperature()'``.

    If the ``name`` of the card is one of the :ref:`Magic Cards <magic-cards>`,
    the value and comment are automatically defined and do not need to be
    specified.

    Parameters
    ----------
    name : str
        The name of the card. Will be trimmed to 8 characters. Can be one of
        the :ref:`Magic Cards <magic-cards>` names, which defines the value
        and comment.
    value
        A Python literal, callable, or string, as described above.
    comment : str
        A short comment describing the card value.
    fargs : tuple
        If ``value`` is a callable, the arguments to pass to it when it gets
        called. The arguments are evaluated following the same rules as with
        the ``value`` before being passed to the callable.
    evaluate : bool
        If `True`, assumes the value is a string to be evaluated in realtime.
        The context is used as ``locals`` for the evaluation. For security,
        ``globals`` are not available.
    context : dict
        A dictionary of parameters used to fill the value replacement fields.
        Two values, ``__exposure__`` and ``__camera__``, are always defined.
        This context can be updated during the evaluation of the card.

    """

    def __init__(self, name, value=None, comment=None, fargs=None,
                 evaluate=False, context=None):

        if isinstance(name, (list, tuple)):
            try:
                name_ = name[0]
                value = name[1]
                comment = name[2]
                fargs = name[3].get('fargs', None)
                context = name[3].get('context', {})
                evaluate = name[3].get('evaluate', False)
            except IndexError:
                pass
            finally:
                name = name_

        self.name = name

        if len(self.name) > 8:
            warnings.warn(f'trimming {self.name} to 8 characters', CardWarning)
            self.name = self.name[:8]

        if self.is_magic():

            if value:
                raise ValueError(f'cannot override value of magic card {self.name!r}')

            magic_params = _MAGIC_CARDS[self.name.upper()]
            self.value, magic_comment = magic_params[0:2]

            self.comment = comment or magic_comment

            if len(magic_params) == 3:
                fargs = magic_params[2].get('fargs', None)
                context = magic_params[2].get('context', {})
                evaluate = magic_params[2].get('evaluate', False)

        else:

            self.value = value
            self.comment = comment

        self._fargs = fargs
        self._evaluate = evaluate

        self.context = context or {}

    def __repr__(self):

        return f'<Card (name={self.name!r}, value={self.value!r})>'

    def is_magic(self):
        """Returns `True` if the card name matches a magic card."""

        return self.name.upper() in _MAGIC_CARDS

    @contextmanager
    def set_exposure(self, exposure, context={}):
        """Sets the current exposure and context, clearing it on exit."""

        original_context = self.context.copy()

        try:
            self._exposure = exposure
            self.context.update({'__exposure__': self._exposure,
                                 '__camera__': self._exposure.camera})
            self.context.update(context)
            yield
        finally:
            self._exposure = None
            self.context = original_context

    def _render_value(self, value):
        """Returns the value after evaluating the string template."""

        if not isinstance(value, str):
            return value

        if self._evaluate:
            return eval(value, {}, self.context)

        # Expand the placeholders. If there are none, no harm done.
        return value.format(**self.context)

    def _evaluate_callable(self, func, use_fargs=True):
        """Evaluates a value that is a callable function or method."""

        if use_fargs and self._fargs:
            # Evaluate the fargs as values so that we can apply
            # all the magic to them as well.
            fargs = (self._evaluate_value(farg, use_fargs=False)
                     for farg in self._fargs)
            return func(*fargs)

        return func()

    def _evaluate_value(self, value, use_fargs=True):
        """Evaluates a value, expanding its template."""

        if isinstance(value, (list, tuple)):
            return (self._evaluate_value(v) for v in value)

        if callable(value):
            return self._evaluate_callable(value, use_fargs=use_fargs)

        return self._render_value(value)

    def evaluate(self, exposure, context={}):
        """Evaluates the card.

        Evaluates the card value, expanding its template parameters. If the
        value is a callable, calls the function at this time.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the card.
        context : dict
            A dictionary of arguments used to evaluate the parameters in
            the value. These argument update those defined when the `.Card`
            was created.

        Returns
        -------
        tuple
            A tuple with the name, evaluated value, and comment for this card.

        """

        with self.set_exposure(exposure, context):
            rendered_value = self._evaluate_value(self.value)

        return (self.name, rendered_value, self.comment)


class CardGroup(list):
    """A group of `.Card` instances.

    A `.CardGroup` is just a list of `Cards <.Card>` that are grouped for
    convenience and for easy reuse.

    Parameters
    ----------
    cards : list
        A list of `.Card` instances. Elements can also be a tuple of two or
        three elements (for name, value, and optionally a comment).
    name : str
        A name for the card group.
    use_group_title : bool
        Whether to prepend a COMMENT card with the group title when creating
        a header.

    """

    def __init__(self, cards, name=None, use_group_title=True):

        self.name = name
        self.use_group_title = use_group_title

        cards = [self._process_input(card) for card in cards or []]
        list.__init__(self, cards)

    def __repr__(self):

        return f'<CardGroup {list.__repr__(self)!s}>'

    def _process_input(self, card):
        """Processes the input and converts it into a valid card."""

        if isinstance(card, Card):
            return card
        elif isinstance(card, (list, tuple, str)):
            return Card(card)
        else:
            raise CardError(f'invalid card {card}')

    def append(self, card):
        list.append(self, self._process_input(card))

    def insert(self, idx, card):
        list.insert(self, idx, self._process_input(card))

    def evaluate(self, exposure, context={}):
        """Evaluates all the cards.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the cards.
        context : dict
            A dictionary of arguments used to evaluate the cards.

        Returns
        -------
        list
            A list of tuples with the name, evaluated value, and comment
            for each card.

        """

        return [card.evaluate(exposure, context=context) for card in self]

    def to_header(self, exposure, context={}, use_group_title=None):
        """Evaluates all the cards and returns a header.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the cards.
        context : dict
            A dictionary of arguments used to evaluate the cards.
        use_group_title : bool
            Whether to prepend a COMMENT card with the group title when
            creating the header.

        Returns
        -------
        header : `~astropy.io.fits.Header`
            A header composed from the cards in the group.

        """

        header = astropy.io.fits.Header(self.evaluate(exposure, context=context))

        use_group_title = use_group_title or self.use_group_title
        if use_group_title and self.name:
            header.insert(0, ('COMMENT', '{s:#^30}'.format(s=f' {self.name} ')))

        return header


class MacroCard(object, metaclass=abc.ABCMeta):
    """A macro that returns a list of keywords and values.

    This is an abstract class in which the `.macro` method must be overridden
    to return a list of keywords and values that can be used to create a
    header.

    Parameters
    ----------
    name : str
        A name for the macro group.
    use_group_title : bool
        Whether to prepend a COMMENT card with the macro title when creating
        a header.
    kwargs : dict
        Additional arguments for the macro.

    """

    def __init__(self, name=None, use_group_title=True, **kwargs):

        self.name = name
        self.use_group_title = use_group_title
        self.kwargs = kwargs

    def __repr__(self):

        return f'<{self.__class__.__name__ (name=self.name)}>'

    @abc.abstractmethod
    def macro(self, exposure, context={}):
        """The macro.

        Must return a list of tuples with the format
        ``(keyword, value, comment)`` or ``(keyword, value)``.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the macro.
        context : dict
            A dictionary of parameters that can be used by the macro.

        """

        raise NotImplementedError

    def evaluate(self, exposure, context={}):
        """Evaluates the macro. Equivalent to calling `.macro` directly.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the macro.
        context : dict
            A dictionary of parameters that can be used by the macro.

        Returns
        -------
        cards : `list`
            A list of tuples with the format ``(keyword, value, comment)``
            or ``(keyword, value)``.

        """

        return self.macro(exposure, context=context)

    def to_header(self, exposure, context={}, use_group_title=None):
        """Evaluates the macro and returns a header.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the macro.
        context : dict
            A dictionary of arguments used to evaluate the macro.
        use_group_title : bool
            Whether to prepend a COMMENT card with the macro title when
            creating the header.

        Returns
        -------
        header : `~astropy.io.fits.Header`
            A header composed from the cards in the macro.

        """

        header = astropy.io.fits.Header(self.evaluate(exposure, context=context))

        use_group_title = use_group_title or self.use_group_title
        if use_group_title and self.name:
            header.insert(0, ('COMMENT', '{s:#^30}'.format(s=f' {self.name} ')))

        return header
